//
//  arifmstream.hpp - Buffered arifmetic encoder and decoder.
//
//  It is file 'arifmstream.hpp'
//
//  Copyright (C) 1992, 1993 Serge Pachkovsky, ps@che.nsk.su
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 1, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//
//  End-user classes, ArifmeticEncode and ArifmeticDecode, defined
//  in this module, are abstract classes - you will need to supply
//  functions WriteBuffer and ReadBuffer to make 'em living.
//


const	int		BitsInByte		= 8 ;
const	unsigned char	TopBitInLatch		= 1 << (BitsInByte - 1) ;
const	int		MaximumArifmeticScale	= 16383 ;
const	int		DefaultBitStreamBuffer	= 2 * 1024 ;

typedef unsigned short	UINT16 ;
typedef short		INT16 ;

enum Boolean {
    False = 0,
    True  = 1
    } ;

class BitStream {
    //
    //	Generic bit stream class
    //
    protected:
    Boolean	    error_flag ;
    unsigned char   *buf ;
    unsigned char   *data_ptr ;
    unsigned char   *end_ptr ;
    size_t	    size ;
    unsigned char   bit_latch ;
    int 	    bits_used_in_latch ;
    public:
		    BitStream( unsigned need_size = DefaultBitStreamBuffer ) ;
		   ~BitStream( void ) ;
    Boolean	    IsError( void ) ;
    } ;

class OutputBitStream : public BitStream {
    //
    //	Abstract output bit stream. WriteBuffer will be called to perform
    //	actual I/O
    //
    unsigned long   StreamSize ;
    void	    FlushOutputByte( void ) ;
    protected:
    virtual size_t  WriteBuffer( unsigned char *, size_t ) = 0 ;
    public:
		    OutputBitStream( void ) ;
		   ~OutputBitStream( void ) ;
    void	    WriteBit( int bit_value ) ;
    void	    FlushOutputBuffer( void ) ;
    unsigned long   GetStreamSize( void ) ;
    } ;

class InputBitStream : public BitStream {
    //
    //	Abstract input bit stream. WriteBuffer will be called to perform
    //	actual I/O
    //
    Boolean	    eof_flag ;
    void	    LoadByteLatch( void ) ;
    void	    LoadInputBuffer() ;
    protected:
    virtual size_t  ReadBuffer( unsigned char *, size_t ) = 0 ;
    public:
		    InputBitStream( void ) ;
    Boolean	    IsEOF( void ){ return eof_flag ; }
    int 	    ReadBit( void ) ;
    } ;

struct Symbol {
    UINT16	    low_count ;
    UINT16	    high_count ;
    UINT16	    scale ;
    } ;

class ArifmeticEncode : public OutputBitStream {
    #ifdef MIRROR
    FILE	   *mirror ;
    #endif
    //
    //	Abstract arifmetic encoder class
    //
    UINT16	    code ;		/* The present input code value       */
    UINT16	    low ;		/* Start of the current code range    */
    UINT16	    high ;		/* End of the current code range      */
    long	    underflow_bits;	/* Number of underflow bits pending   */
    public:
		    ArifmeticEncode( void ) ;
		   ~ArifmeticEncode( void ) ;
    void	    EncodeSymbol( const Symbol &s ) ;
    void	    FlushOutputBuffer( void ) ;
    } ;

class ArifmeticDecode : public InputBitStream {
    //
    //	Abstract arifmetic decoder class
    //
    UINT16	    code ;		/* The present input code value       */
    UINT16	    low ;		/* Start of the current code range    */
    UINT16	    high ;		/* End of the current code range      */
    long	    underflow_bits;	/* Number of underflow bits pending   */
    public:
    void	    StartDecoder( void ) ;
    UINT16	    DecodeSymbol( const Symbol &s ) ;
    void	    RemoveSymbol( const Symbol &s ) ;
    } ;

//
//  Class member are all inline - they will be called really often...
//

inline
BitStream::BitStream( unsigned need_size )
{
error_flag	   = False ;
size		   = need_size ;
buf		   = new unsigned char [ size ] ;
data_ptr	   = buf ;
end_ptr 	   = buf + size ;
bits_used_in_latch = 0 ;

if( buf == NULL ){
    cerr << "Error initializing bit stream - out of memory\n" ;
    exit( 3 ) ;
    }
}

inline
BitStream::~BitStream( void )
{
delete buf ;
}

inline Boolean 
BitStream::IsError( void )
{
return error_flag ;
}

inline
OutputBitStream::OutputBitStream( void )
{
StreamSize = 0 ;
}

inline
OutputBitStream::~OutputBitStream( void )
{
if( error_flag )
    return ;
}

inline void
OutputBitStream::FlushOutputByte( void )
{
*data_ptr++	   = bit_latch ;
StreamSize++ ;
bits_used_in_latch = 0 ;
if( data_ptr >= end_ptr )
    FlushOutputBuffer() ;
}

inline unsigned long
OutputBitStream::GetStreamSize( void )
{
return StreamSize ;
}

inline void
OutputBitStream::WriteBit( int bit_value )
{
bit_latch = ( bit_latch << 1 ) | ( bit_value != 0 ) ;
if( ++bits_used_in_latch >= BitsInByte )
    FlushOutputByte() ;
}

#ifdef GEN_HUGE_FUNCS
void HugeInline
OutputBitStream::FlushOutputBuffer( void )
{
if( bits_used_in_latch > 0 ){
    //
    //	Note what this code will never execute when FlushOutputBuffer was
    //	called from FlushOutputByte, because byte latch was already flushed
    //	in that case.
    //
    for( int i = bits_used_in_latch ; i < BitsInByte ; i++ )
	WriteBit( 0 ) ;
    }
    size_t  xfer_size = data_ptr - buf ;

if( WriteBuffer( buf, xfer_size ) != xfer_size ){
    cerr << "Out of disk space writing to the output stream!\n" ;
    exit( -1 ) ;
    error_flag = True ;
    }
data_ptr = buf ;
}
#endif

inline
InputBitStream::InputBitStream( void )
{
data_ptr    = buf + size ;
eof_flag    = False ;
}

#ifdef GEN_HUGE_FUNCS
void HugeInline
InputBitStream::LoadInputBuffer( void )
{
    size_t  xfer_size ;

xfer_size = ReadBuffer( buf, size ) ;
if( xfer_size == (size_t)-1 ){
    error_flag = True ;
    xfer_size  = 0 ;
    cerr << "InputBitStream::LoadInputBuffer: error reading compressed file\n" ;
    exit( -1 ) ;
    }
if( xfer_size == 0 ){
    eof_flag   = True ;
    }
data_ptr       = buf ;
end_ptr        = buf + xfer_size ;
}
#endif

inline void
InputBitStream::LoadByteLatch( void )
{
if( data_ptr >= end_ptr )
    LoadInputBuffer() ;
if( data_ptr >= end_ptr ){
    eof_flag = True ;
    return ;
    }
bit_latch	   = *data_ptr++ ;
bits_used_in_latch = BitsInByte ;
}

inline int
InputBitStream::ReadBit( void )
{
if( bits_used_in_latch <= 0 )
    LoadByteLatch() ;
if( IsEOF() ){
    return 0 ;
    }
bits_used_in_latch-- ;
if( bit_latch & TopBitInLatch ){
    bit_latch <<= 1 ;
    return 1 ;
    }
else {
    bit_latch <<= 1 ;
    return 0 ;
    }
}

inline
ArifmeticEncode::ArifmeticEncode( void )
{
low	       = 0;
high	       = 0xffff;
underflow_bits = 0;
#ifdef MIRROR
mirror	       = fopen( "mirror.$1", "wb" ) ;
#endif
}

inline
ArifmeticEncode::~ArifmeticEncode( void )
{
#ifdef MIRROR
fclose( mirror ) ;
#endif
}

inline void
ArifmeticEncode::FlushOutputBuffer( void )
{
WriteBit( low & 0x4000 ) ;
underflow_bits++ ;
while( underflow_bits-- > 0 )
    WriteBit( ~low & 0x4000 ) ;
OutputBitStream::FlushOutputBuffer() ;
}

inline void
ArifmeticDecode::StartDecoder( void )
{
    int     i ;

for( code = 0, i = 0 ; i < 16 ; i++ )
    code = ( code << 1 ) | ReadBit() ;
low  = 0 ;
high = 0xffff ;
}

#ifdef GEN_HUGE_FUNCS
void HugeInline
ArifmeticEncode::EncodeSymbol( const Symbol &s )
{
    UINT16	    t_range ;	// This is necessary to convince GCC 1.39
				// not to promote UINT16 to unsigned long...
    unsigned long   range ;

#ifdef	DEBUG
//
//  Few sanity checks
//
if( s.low_count > s.scale || s.high_count > s.scale || s.low_count >= s.high_count )
    cerr << "ArifmeticEncode::EncodeSymbol("
	 << s.low_count << ","
	 << s.high_count << ","
	 << s.scale << ") sanity check failed\n" ;
#endif
#ifdef MIRROR
fwrite( &s, sizeof s, 1, mirror ) ;
#endif
//
// These three lines rescale high and low for the new symbol.
//
t_range = high - low ;
range	= (unsigned long) t_range + 1 ;
#ifdef DEBUG
if( range == 0 || range > 0x10000L ){
    cerr << "ArifmeticEncode::EncodeSymbol: impossible range " << range << "\n" ;
    exit( -1 ) ;
    }
#endif
high	= low + UINT16( ( range * s.high_count ) / s.scale - 1 ) ;
low	= low + UINT16( ( range * s.low_count  ) / s.scale ) ;
//
// This loop turns out new bits until high and low are far enough
// apart to have stabilized.
//
for( ; ; low <<= 1, high = ( high << 1 ) | 1 ){
    //
    // If this test passes, it means that the MSDigits match, and can
    // be sent to the output stream.
    //
    if( ( (high ^ low) & 0x8000 ) == 0 ){
	WriteBit( high & 0x8000 ) ;
	for( ; underflow_bits > 0 ; underflow_bits-- )
	    WriteBit( ~high & 0x8000 ) ;
	}
    //
    // If this test passes, the numbers are in danger of underflow, because
    // the MSDigits don't match, and the 2nd digits are just one apart.
    //
    else
    if( ( low & 0x4000 ) && !( high & 0x4000 ) ){
	underflow_bits++ ;
	low  &= 0x3fff ;
	high |= 0x4000 ;
	}
    else
	break ;
    }
#ifdef MIRROR
fwrite( &high, sizeof high, 1, mirror ) ;
fwrite( &low,  sizeof low,  1, mirror ) ;
fwrite( &underflow_bits, sizeof underflow_bits, 1, mirror ) ;
#endif
}
#endif

//
// When decoding, this routine is called to figure out which symbol
// is presently waiting to be decoded.	This routine expects to get
// the current model scale in the s->scale parameter, and it returns
// a count that corresponds to the present floating point code:
//
//  code = count / s->scale
//
inline UINT16 
ArifmeticDecode::DecodeSymbol( const Symbol &s )
{
    UINT16  t_range1 = code - low ;
    UINT16  t_range2 = high - low ;
    UINT16  result ;

result = UINT16( ( ( (unsigned long)( t_range1 ) + 1 ) * s.scale - 1 ) /
		   ( (unsigned long)( t_range2 ) + 1 ) ) ;
#ifdef	DEBUG
//
//  Sanity checks
//
if( result > s.scale )
    cerr << "ArifmeticDecode::DecodeSymbol(" << s.scale << ") sanity check: got " << result << "\n" ;
#endif
return result ;
}

//
// Just figuring out what the present symbol is doesn't remove
// it from the input bit stream.  After the character has been
// decoded, this routine has to be called to remove it from the
// input stream.
//
inline void
ArifmeticDecode::RemoveSymbol( const Symbol &s )
{
    UINT16	    t_range ;	// This is necessary to convince GCC 1.39
				// not to promote UINT16 to unsigned long...
    unsigned long   range ;

#ifdef	DEBUG
//
//  Few sanity checks
//
if( s.low_count > s.scale || s.high_count > s.scale || s.low_count >= s.high_count )
    cerr << "ArifmeticEncode::RemoveSymbol("
	 << s.low_count << ","
	 << s.high_count << ","
	 << s.scale << ") sanity check failed\n" ;
#endif
//
// First, the range is expanded to account for the symbol removal
//
t_range = high - low ;
range	= (unsigned long) t_range + 1 ;
#ifdef DEBUG
if( range == 0 || range > 0x10000L ){
    cerr << "ArifmeticEncode::EncodeSymbol: impossible range " << range << "\n" ;
    exit( -1 ) ;
    }
#endif
high	= low + UINT16( ( range * s.high_count ) / s.scale - 1 ) ;
low	= low + UINT16( ( range * s.low_count  ) / s.scale     ) ;
//
// Next, any possible bits are shipped out.
//
for( ; ; low <<= 1, high = ( high << 1 ) | 1, code = ( code << 1 ) | ReadBit() ){
    //
    // If the MSDigits match, the bits will be shifted out.
    //
    if( ( (high ^ low) & 0x8000 ) == 0 ){
	}
    //
    // Else, if underflow is threatining, shift out the 2nd MSDigit.
    //
    else
    if( (low & 0x4000) && !(high & 0x4000) ){
	code ^= 0x4000;
	low  &= 0x3fff;
	high |= 0x4000;
	}
    //
    // Otherwise, nothing can be shifted out, so I return.
    //
    else
	break ;
    }
}

