//
//  It is file 'codes.hpp'
//
//  Copyright (C) 1992, 1993 Serge Pachkovsky, ps@che.nsk.su
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 1, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//
//  This code is based on ideas found in Mark Nelson's code (DDJ 91/02,
//  article on Arithmetic coding and statistical modelling).
//
//  The only two end-user classes in this file are
//  Model2Compressor and Model2Decompressor, which are
//  quite simple and pleasant to use, unlike underlying
//  ones...
//

//
//  Originally, I decided to flush model by sorting contexts at the
//  each level by their's use counts, deleting fixed portion of ones
//  which are the least used. This proven to bee _too_ slow with
//  great number of contexts, so I decided simply to divide use counts
//  by a predefined constant, removing contexts with use counts fallen
//  to zero. If you wish to use original flushing strategy, please
//  comment out the following string.
//
#define FAST_FLUSHING

//
//  In theory, loose-array representation of character counts could
//  sometimes (when probabilities of characters are distributed almost
//  uniformly) cause memory loss. To account for such cases, I introduced
//  the second context format, with stats in ordinary array format.
//  Handling it required a bit more time in Totals::Fill, and, as the
//  number of such contexts is usually quite small (say, 50 out of 10000),
//  I will put code responsible forhandling it inside ifdef's....
//
#define NO_FAST_STATS

//
//  Enabling this define will cause compressor to "burst" arriving codes,
//  i.e., then code appears first time, it is given count CodeBurstValue
//  instead of 1. If the very next symbol appearing in this context is the
//  same, bursted count is approved. If no, burst value is discarded...
//  Looks like it helps to increase compression ratio if burst procedure
//  is applied to leaf nodes only...
//  This should help (I hope so) to compress binary files, which are the
//  weak spot of PSA... sorry, no way :(
//
#define CODE_BURST
const	int	CodeBurstValue	= 2 ;


#if defined( _MSC_VER )
#   pragma inline_depth( 255 )
#   pragma inline_recursion( on )
#endif

#if defined( __TURBOC__ ) || defined( _MSC_VER )
#   define  MSDOS
#endif

#ifdef	DEBUG
void DPrint( char *msg ) ;
#else
#define DPrint(s)
#endif

#ifdef	DEBUG
#define REPORT_FLUSHING
#endif

#ifdef	DEBUG
void inline
DPrint( char *msg )
{
cerr << msg ;
}

int inline
SanePointer( const void *const ptr )
{
#ifdef MSDOS
#define FP_SEG(p)   ((unsigned long)(p)>>16)
if( ptr == NULL || ( FP_SEG( ptr ) > _CS && FP_SEG( ptr ) < 0xA000 ) )
    return 1 ;
return 0 ;
#undef	FP_SEG
#else
//
//	This assumes flat memory, with code at the lower addresses and 
//	memory arena growing upwards...
//
if( ptr == NULL ){
	cout << "Pointer " << ptr << " is zero\n" ;
	return 0 ;
	}
if( ptr > 0x10800000 ){ // I should have used sbrk(0) instead, but that's too slow...
	cout << "Pointer " << ptr << " is above memory arena\n" ;
	return 0 ;
	}
if( ptr <= (void *) SanePointer ){
	cout << "Pointer " << ptr << " is in code region\n" ;
	return 0 ;
	}
return 1 ;
#endif
}
#endif

//
//  Some magic types and constants
//
const	int		MaxModelOrder		= 25 ;
const	int		CharsCount		= 256 ;
const	int		CountTypeMax		= 255 ;
typedef unsigned char	CountType ;
typedef unsigned char	CharType ;
const	unsigned long	DefaultMemoryLimit	= 400 * 1024 ;
const	int		DefaultOrder		= 4 ;

//
//  Here are memory-control variables.
//
//  LowModelMemory and LowModelMemoryPart both control the model
//  flushing strategy. Each time the available memory falls down
//  to LowModelMemory or total memory/LowModelMemoryPart (whichever
//  is bigger), model flushing is performed. Although setting
//  LowModelMemoryPart will result in better compression ratiou,
//  it will cause garbage collection to run more often, thus hurting
//  performance.
//
const	unsigned long	LowModelMemory		= 5 * 1024 ;
const	int		LowModelMemoryPart	= 20 ;
//
//  These constants control details of model flushing.
//
//  SacredLinks is the minimal number of links which should never
//  be flushed from context. Normally, it should be zero.
//  LeastUsedCutoffPart is the portion of context's links which
//  should be removed at each model flushing. Setting it to the
//  higher value will increase compression at the expense of
//  performance. Contexts use counts are scaled down by
//  ModelWeightsReduce at each flushing.
//
const	int		SacredLinks		= 0 ;
const	int		LeastUsedCutoffPart	= 16 ;
const	int		ModelWeightsReduce	= 2 ;

//
//  Compression classes interfaces definitions
//

const int	    SpecialOrders   = 2 ;
//
//  Special "action" codes:
//
const int	    Escape	    = -1 ;
const int	    Done	    = -2 ;
const int	    Flush	    = -3 ;

#include "arifmstr.hpp"
#include "memory.hpp"
#include "scoreboard.hpp"
#include "cpath.hpp"
#include "tot_dcl.hpp"
#include "con_dcl.hpp"

class Model2 {
    static Model2  *active_model ;
    MemoryAllocator memory ;
    int 	    max_order ;
    int 	    current_order ;
    #ifdef CODE_BURST
    int 	    burstable_order ;
    #endif
    Context	   *contexts  [ MaxModelOrder + SpecialOrders + 1 ] ;
    ContextPath     path ;
    Totals	    totals ;
    int 	    codes_count ;
    Context *	   &Contexts( int order ) ;
    void	    SetupHighOrderContexts( void ) ;
    Context *	    MakePathContext( int order ) ;
    void	    ShiftToNextContext( CharType ) ;
    void	    RecursiveFlushContexts( Context * ) ;
    #ifdef STATS
    unsigned long   Flushes ;
    unsigned long   Rescalings ;
    unsigned long   CharCountOverflows ;
    unsigned long   TotalCountOverflows ;
    unsigned long   Encodings ;
    unsigned long   Escapes ;
    unsigned long   ToFastConversions ;
    unsigned long   ToSlowConversions ;
    unsigned long   HitThisContext[ MaxModelOrder + SpecialOrders + 1 ] ;
    unsigned long  &HitCount( int order ) ;
    #endif
    //
    //	Memory allocation interface - I am far from being happy with
    //	standart one - it wastes at least half of the memory I request...
    //
    //	And, these new allocators can't return NULL - if they fail,
    //	they will abort process. It makes my life much simpler...
    //
    friend class	Context ;
    #ifdef STATS
    friend class	MemoryAllocator ;
    friend class	Totals ;
    #endif
    void		SetActiveModel( void ) ;
    //
    //	And specific allocation functions - they will call constructors
    //	and destructors properly...
    //
    static Stats *	NewStats( int, Stats ** ) ;
    static void 	DeleteStats( Stats * ) ;
    static FStats *	NewFastStats( FStats ** ) ;
    static void 	DeleteFastStats( FStats * ) ;
    static Links *	NewLinks( int, Links ** ) ;
    static void 	DeleteLinks( Links * ) ;

    public:
		    Model2( int, unsigned long, int = CharsCount ) ;
		   ~Model2( void ) ;
    unsigned long   GetMemoryUsed( void ) ;
    unsigned long   GetMemoryFree( void ) ;
    unsigned long   GetMaxMemoryUsed( void ) ;
    void	    UpdateModel( int ) ;
    Boolean	    EncodeInt( int, Symbol & ) ;
    void	    GetCurrentScale( Symbol & ) ;
    int 	    DecodeInt( int, Symbol & ) ;
    void	    FlushModel( void ) ;
    } ;

#include "tot_def.hpp"
#include "con_def.hpp"

class Model2Compressor : public ArifmeticEncode {
#ifdef MIRROR
    FILE	   *mirror ;
#endif
    Model2	    model ;
    unsigned long   flush_threshold ;
    int 	    handle ;
    virtual size_t  WriteBuffer( unsigned char *, size_t ) ;
    void	    EncodeInt( int c ) ;
    public:
		    Model2Compressor( int, int = DefaultOrder, unsigned long = DefaultMemoryLimit, int = CharsCount ) ;
		   ~Model2Compressor( void ) ;
    unsigned long   GetMaxMemoryUsed() ;
    void	    WriteChar( CharType ) ;
    } ;

class Model2Decompressor : public ArifmeticDecode {
    Model2	    model ;
    int 	    handle ;
    virtual size_t  ReadBuffer( unsigned char *, size_t ) ;
    int 	    DecodeInt( void ) ;
    public:
		    Model2Decompressor( int, int = DefaultOrder, unsigned long = DefaultMemoryLimit, int = CharsCount ) ;
		   ~Model2Decompressor( void ) ;
    int 	    ReadChar( void ) ;
    } ;

//
//  I made all class members inline functions, so what no
//  efficiency loss will occur due to function call overhead...
//  When using 2M+ of dynamic data space, few dozen Kbytes of
//  code do not matter much...
//
//  Well, GCC 1.39 will pick functions for inline expansion _very_
//  strangely if they are called from relatively big inline
//  function. So, I have to
//

inline Stats *
Model2::NewStats( int count, Stats **ptr )
{
return (Stats *)active_model->memory.AllocateMoveable( sizeof( Stats ) * count, (void **)ptr ) ;
}

inline void
Model2::DeleteStats( Stats *p )
{
active_model->memory.FreeMoveable( p ) ;
}

inline FStats *
Model2::NewFastStats( FStats **ptr )
{
return (FStats *)active_model->memory.AllocateMoveable( sizeof( FStats ) * CharsCount, (void **)ptr ) ;
}

inline void
Model2::DeleteFastStats( FStats *p )
{
active_model->memory.FreeMoveable( p ) ;
}

inline Links *
Model2::NewLinks( int count, Links **ptr )
{
return (Links *)active_model->memory.AllocateMoveable( sizeof( Links ) * count, (void **)ptr ) ;
}

inline void
Model2::DeleteLinks( Links *p )
{
active_model->memory.FreeMoveable( p ) ;
}

inline void
Model2::SetActiveModel( void )
{
active_model = this ;
}

inline Context * &
Model2::Contexts( int order )
{
#ifdef DEBUG
if( order > max_order ){
    cerr << "Model2::Contexts sanity check: order " << order << " requested\n" ;
    exit( -1 ) ;
    }
#endif
return contexts[ order + SpecialOrders ] ;
}

#ifdef STATS
inline unsigned long &
Model2::HitCount( int order )
{
return HitThisContext[ order + SpecialOrders ] ;
}
#endif

#ifdef GEN_HUGE_FUNCS
HugeInline
Model2::Model2( int order, unsigned long max_memory, int codes ) :
    memory( max_memory, sizeof( Context ) )
{
#ifdef STATS
Rescalings	    =
ToFastConversions   =
ToSlowConversions   =
Flushes 	    =
CharCountOverflows  =
TotalCountOverflows =
Encodings	    =
Escapes 	    = 0 ;
memset( HitThisContext, 0, sizeof HitThisContext ) ;
#endif
SetActiveModel() ;

    int 	i ;

if( order > MaxModelOrder ){
    cerr << "Can't initialize model for too big order " << order << "\n" ;
    exit( 3 ) ;
    }
codes_count	= codes ;
current_order	=
max_order	= order ;
#ifdef CODE_BURST
burstable_order = order ;
#endif
//
//  Special codes context
//
Contexts( -2 ) = new Context( Leaf ) ;
Contexts( -2 )->IncrementCharCount( CharType( -Done  ) ) ;
Contexts( -2 )->IncrementCharCount( CharType( -Flush ) ) ;
//
//  Catch-all context for unseen characters
//
Contexts( -1 ) = new Context( Leaf ) ;
for( i = 0 ; i < CharsCount ; i++ )
    Contexts( -1 )->IncrementCharCount( CharType( i ) ) ;
//
//  Create constant zero-order context
//
Contexts(  0 ) = new Context( max_order > 0 ? Node : Leaf ) ;
//
//  Initialize higher-order contexts
//
SetupHighOrderContexts() ;
}

HugeInline
Model2::~Model2( void )
{
SetActiveModel() ;
//
//  Delete special-purpose contexts
//
delete Contexts( -2 ) ;
delete Contexts( -1 ) ;
//
//  And recursively delete all other contexts
//
delete Contexts(  0 ) ;
#ifdef STATS
cerr.flush() ;
cout << "\n     Compression model statistics:\n" ;
cout << "Model flushes                           = " << Flushes << "\n" ;
cout << "Total rescalings                        = " << Rescalings << "\n" ;
cout << "Rescalings due to single count overflow = " << CharCountOverflows << "\n" ;
cout << "Rescalings due to total count overflow  = " << TotalCountOverflows << "\n" ;
cout << "Conversions to fast stats format        = " << ToFastConversions << "\n" ;
cout << "Conversions to slow stats format        = " << ToSlowConversions << "\n" ;
cout << "Total encodings                         = " << Encodings << "\n" ;
cout << "Escapes                                 = " << Escapes << "\n" ;
for( int i = -2 ; i <= max_order ; i++ ){
    cout << "Encodings ended in context " ;
    cout << setw( 3 ) << i ;
    cout << " = " << HitCount( i ) << "\n" ;
    }
cout.flush() ;
#endif
}

HugeInline Context *
Model2::MakePathContext( int order )
{
    Context    *parent = Contexts( 0 ) ;
    Context    *child ;
    int 	t_order ;
    CharType	c ;

for( t_order = 1 ; t_order <= order ; t_order++, parent = child ){
    c = path[ order - t_order ] ;
    if( ( child = parent->GetNext( c ) ) == NULL ){
	child = new Context( t_order < max_order ? Node : Leaf ) ;
	#ifdef CODE_BURST
	if( t_order >= burstable_order )
	    child->SetBurstableFlag() ;
	#endif
	#ifdef DEBUG
	child->SetContextId( t_order, parent, c ) ;
	#endif
	parent->SetNext( c, child ) ;
	}
    }
#ifdef DEBUG
for( int i = 0 ; i < order ; i++ )
    if( path[ i ] != parent->context_id[ order - 1 - i ] ){
	cerr << "Model2::MakePathContext: Context path differs from the model's one\n" ;
	exit( -1 ) ;
	}
#endif
return parent ;
}

#endif

inline void
Model2::SetupHighOrderContexts( void )
{
    int 	order ;

for( order = 1 ; order <= max_order ; order++ )
    Contexts( order ) = MakePathContext( order ) ;
}

#ifdef GEN_HUGE_FUNCS
HugeInline void
Model2::ShiftToNextContext( CharType c )
{
    int 	order ;
    Context    *lesser, *next ;

for( order = max_order ; order > 0 ; order-- ){
    lesser = Contexts( order - 1 ) ;
    if( ( next = lesser->GetNext( c ) ) == NULL ){
	next = new Context( order < max_order ? Node : Leaf ) ;
	#ifdef CODE_BURST
	if( order >= burstable_order )
	    next->SetBurstableFlag() ;
	#endif
	#ifdef DEBUG
	next->SetContextId( order, lesser, c ) ;
	#endif
	lesser->SetNext( c, next ) ;
	}
    Contexts( order ) = next ;
    }
path << c ;
}
#endif

#ifdef GEN_HUGE_FUNCS
HugeInline void
Model2::UpdateModel( int c )
{
SetActiveModel() ;
//
//  Don't need to update model for a special symbols,
//  just get ready for next symbol.
//
if( c >= 0 ){
    int 	order ;
    //
    //	Update counters for the orders used in encoding,
    //	when switch to the new context...
    //
    for( order = 0 ; order < current_order ; order++ )
	Contexts( order )->IncrementUseCount() ;

    for( ; order <= max_order ; order++ ){
	Contexts( order )->IncrementUseCount() ;
	Contexts( order )->IncrementCharCount( CharType( c ) ) ;
	}

    ShiftToNextContext( CharType( c ) ) ;
    }
current_order = max_order ;
totals.ClearScores() ;
}
#endif

#ifdef GEN_HUGE_FUNCS
HugeInline Boolean
Model2::EncodeInt( int c, Symbol &code )
{
SetActiveModel() ;

    Context	*context = Contexts( current_order ) ;
    int 	index ;

#ifdef DEBUG
if( current_order < -2 )
    cerr << "Model2::EncodeInt: escaped to context " << current_order << "?!\n" ;
#endif
totals.Fill( context, current_order == -2 ? True : False, codes_count ) ;
do {
    if( (c < 0) + (current_order <= -2) == 1 ){
	//
	//  Values lesser then zero could only be found in
	//  context -2
	//
	current_order-- ;
	index = -1 ;
	break ;
	}
    if( current_order == -2 )
	c = -c ;
    if( ( index = context->CharToIndex( CharType( c ) ) ) < 0 || context->GetCount( index ) == 0 ){
	current_order-- ;
	index = -1 ;
	}
    } while( 0 ) ;
code.scale	= totals.GetScale() ;
code.low_count	= totals.LowCount ( index ) ;
code.high_count = totals.HighCount( index ) ;
#ifdef DEBUG
if( code.low_count >= code.high_count || code.high_count > code.scale )
    cerr << "Model2::EncodeInt: invalid encoding (" << code.low_count <<
	    "," << code.high_count << "," << code.scale << ")\n" ;
#endif
#ifdef TRACE_CODES
cout << "(" << code.low_count << "," << code.high_count << "," << code.scale << ")\n" ;
#endif
#ifdef STATS
Encodings ++ ;
if( index < 0 )
     Escapes ++ ;
else HitCount( current_order ) ++ ;
#endif
return index < 0 ? True : False ;
}
#endif

inline void
Model2::GetCurrentScale( Symbol &code )
{
SetActiveModel() ;
totals.Fill( Contexts( current_order ), current_order == -2 ? True : False, codes_count ) ;
code.scale = totals.GetScale() ;
}

inline int
Model2::DecodeInt( int count, Symbol &code )
{
SetActiveModel() ;

    int 	index = totals.SearchCount( count ) ;

#ifdef DEBUG
if( current_order < -2 )
    cerr << "Model2::DecodeInt: escaped to context " << current_order << "?!\n" ;
#endif
code.low_count	= totals.LowCount ( index ) ;
code.high_count = totals.HighCount( index ) ;
#ifdef TRACE_CODES
cout << "(" << code.low_count << "," << code.high_count << "," << code.scale << ")\n" ;
#endif
if( index < 0 ){
    current_order-- ;
    return Escape ;
    }
if( current_order >= -1 )
     return  (int)Contexts( current_order )->IndexToChar( index ) ;
else return -(int)Contexts( current_order )->IndexToChar( index ) ;
}

inline unsigned long
Model2::GetMemoryUsed( void )
{
return memory.MemoryUsed() ;
}

inline unsigned long
Model2::GetMemoryFree( void )
{
return memory.MemoryFree() ;
}

inline unsigned long
Model2::GetMaxMemoryUsed()
{
return memory.MaxMemoryUsed() ;
}

#ifdef GEN_HUGE_FUNCS
HugeInline void
Model2::RecursiveFlushContexts( Context *context )
{
#ifdef DEBUG
if( ! SanePointer( context ) )
	cerr << "Model2::RecursiveFlushContexts: Received insane context " << context << "\n" ;
#endif
context->ScaleUseCount() ;
if( context->IsLeaf() )
    return ;

context->DeleteLeastUsed() ;

    int 	    i = context->GetMaxIndex() ;
    Context	   *next ;

for( ; i >= 0 ; i-- )
    if( ( next = context->GetNext( i ) ) != NULL )
	RecursiveFlushContexts( next ) ;
}

HugeInline void
Model2::FlushModel( void )
{
SetActiveModel() ;
#ifdef STATS
Flushes++ ;
#endif
#ifdef REPORT_FLUSHING
cerr << "Flushing (mem = " << GetMemoryUsed() << ") ..." ;
#else
cerr << "$" ; cerr.flush() ;
#endif
RecursiveFlushContexts( Contexts( 0 ) ) ;
//
//  Active contexts could have being purged, let's make sure
//  they are here...
//
SetupHighOrderContexts() ;
#ifdef REPORT_FLUSHING
cerr << "done (mem = " << GetMemoryUsed() << ")\n" ;
#else
cerr << "\b \b" ; cerr.flush() ;
#endif
}

HugeInline
Model2Compressor::Model2Compressor( int h, int order, unsigned long memory, int codes ) :
    model( order, memory, codes )
{
handle		= h ;
flush_threshold = memory / LowModelMemoryPart ;
if( flush_threshold < LowModelMemory )
    flush_threshold = LowModelMemory ;
#ifdef MIRROR
mirror = fopen( "mirror.$", "wb" ) ;
#endif
}

HugeInline
Model2Compressor::~Model2Compressor( void )
{
EncodeInt( Done ) ;
FlushOutputBuffer() ;
#ifdef MIRROR
fclose( mirror ) ;
#endif
}
#endif

inline size_t
Model2Compressor::WriteBuffer( unsigned char *buf, size_t size )
{
return write( handle, buf, size ) ;
}

inline unsigned long
Model2Compressor::GetMaxMemoryUsed()
{
return model.GetMaxMemoryUsed() ;
}

#ifdef GEN_HUGE_FUNCS
HugeInline void
Model2Compressor::EncodeInt( int c )
{
    Boolean	escaped ;
    Symbol	code ;

do {
    escaped = model.EncodeInt( c, code ) ;
    EncodeSymbol( code ) ;
    #ifdef MIRROR
    fwrite( &code, sizeof code, 1, mirror ) ;
    #endif
    } while( escaped ) ;
model.UpdateModel( c ) ;
}
#endif

inline void
Model2Compressor::WriteChar( CharType c )
{
#ifdef MIRROR
putc( c, mirror ) ;
#endif
EncodeInt( int( c ) ) ;
while( model.GetMemoryFree() < flush_threshold ){
    EncodeInt( Flush ) ;
    model.FlushModel() ;
    }
}

#ifdef GEN_HUGE_FUNCS
HugeInline
Model2Decompressor::Model2Decompressor( int h, int order, unsigned long max_memory, int codes ) :
    model( order, max_memory, codes )
{
handle = h ;
StartDecoder() ;
}

HugeInline
Model2Decompressor::~Model2Decompressor( void )
{
}
#endif

inline size_t
Model2Decompressor::ReadBuffer( unsigned char *buf, size_t size )
{
return read( handle, buf, size ) ;
}

#ifdef GEN_HUGE_FUNCS
HugeInline int
Model2Decompressor::DecodeInt( void )
{
    Symbol	code ;
    UINT16	value ;
    int 	c ;

do {
    model.GetCurrentScale( code ) ;
    value = DecodeSymbol( code ) ;
    c	  = model.DecodeInt( value, code ) ;
    #ifdef DEBUG
    if( code.low_count > value || code.high_count <= value )
	cerr << "Character decoded incorrectly: value " << value << " not in range (" << code.low_count << "," << code.high_count << ")\n" ;
    #endif
    RemoveSymbol( code ) ;
    } while( c == Escape ) ;
model.UpdateModel( c ) ;
return c ;
}
#endif

inline int
Model2Decompressor::ReadChar( void )
{
    int 	code ;

do {
    switch( code = DecodeInt() ){
	case Done:
	    return -1 ;
	case Flush:
	    model.FlushModel() ;
	    break ;
	#ifdef DEBUG
	default:
	    if( code < 0 )
		cerr << "Model2Decompressor::ReadChar: undefined special code " << code << " encountered\n" ;
	#endif
	}
    } while( code < 0 ) ;
return code ;
}
