//
//  It is file 'con_def.hpp'
//
//  Copyright (C) 1992, 1993 Serge Pachkovsky, ps@che.nsk.su
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 1, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//
//  Implemetations of members...
//

#ifdef DEBUG
void
Context::DumpContextInfo( void ) const
{
cerr << "Context::DumpContextInfo: " 			<< "\n"
     << "	context_level	= " << context_level    << "\n"
     << "	context_id      = " << context_id    	<< "\n"
     << "	flags		= " << flags    	<< "\n"
     << "	rescale_count	= " << rescale_count    << "\n"
     << "	burst_char	= " << burst_char    	<< "\n"
     << "	max_index       = " << max_index    	<< "\n"
     << "	use_count	= " << use_count    	<< "\n"
     << "	links	        = " << links    	<< "\n"
     << "	stats		= " << stats    	<< "\n"
     << "	fstats		= " << fstats    	<< "\n"
     ;
}
#endif

inline
Context::Context( ContextType tp )
{
flags	      = 0 ;
if( tp == Leaf )
    flags |= ContextLeaf ;
max_index     = -1 ;
stats	      = NULL ;
links	      = NULL ;
use_count     = 0 ;
rescale_count = 0 ;
#ifdef DEBUG
context_level = 0 ;
#endif
#ifdef CODE_BURST
burst_char    = NoBurstChar ;
#endif
}

#ifdef DEBUG
#ifdef GEN_HUGE_FUNCS
HugeInline void
Context::SetContextId( int level, Context *parent, CharType c )
{
if( parent->context_level != level - 1 ){
    cerr << "Context::SetContextId: invalid context linkage: order " << level << " linked to " << parent->context_level << "\n" ;
    exit( -1 ) ;
    }
if( parent->context_level > 0 )
    memcpy( context_id, parent->context_id, parent->context_level ) ;
if( level > 0 )
    context_id[ level - 1 ] = c ;
context_level = level ;
}
#endif
#endif

inline
Context::~Context( void )
{
    int 	i, size ;
    Context	*next ;

#ifdef DEBUG
CheckSanity( "Destructor" ) ;
#endif
if( ! IsLeaf() ){
    //
    //	Kill all sublings...
    //
    for( i = 0, size = GetMaxIndex() ; i <= size ; i++ )
	if( ( next = GetNext( i ) ) != NULL )
	    delete next ;
    Model2::DeleteLinks( links ) ;
    }
if( FastStats() )
     Model2::DeleteFastStats( fstats ) ;
else Model2::DeleteStats( stats ) ;
#ifdef PATCH_DELETE_BUG
//
//  G++ 1.39 have a horrible bug in operator delete() handling:
//  for all classes with own operator delete(), it is called with
//  NULL argument, regardless of address of object being deleting.
//  Therefore, we will call operator delete() by hand...
//
Context::operator delete( this ) ;
#endif
}

#ifdef DEBUG
inline void
Context::CheckSanity( const char *from ) const
{
	const char 	*msg ;

if( ! SanePointer( this ) ){
    cerr << "Context::CheckSanity: Insane context pointer " << (void *)this << "\n" ;
    cerr << "Called from: " << from << "\n" ;
    exit( -1 ) ;
    }
if( ( msg = "max_index >= CharsCount", max_index >= CharsCount ) || 
    ( msg = "invalid bits in flags set", (flags & ~ContextAllFlags ) != 0 ) ||
    ( msg = "insane stats structure pointer", stats != NULL && ! SanePointer( stats ) ) || 
    ( msg = "insane links structure pointer", links != NULL && ! SanePointer( links ) ) ){
    cerr << "Context::CheckSanity: Insane context data fields: " << msg << "\n" ;
    cerr << "Called from: " << from << "\n" ;
    exit( -1 ) ;
    }
}
#endif

inline Boolean
Context::FastStats( void ) const
{
#ifdef NO_FAST_STATS
return False ;
#else
return flags & ContextFast ? True : False ;
#endif
}

inline Boolean
Context::GoodForLoose( int size )
{
#ifdef NO_FAST_STATS
return True ;
#else
return Boolean( size < ( IsLeaf() ? MaxLooseLeafLinks : MaxLooseNodeLinks ) ) ;
#endif
}

inline int
Context::GetRescaleCount() const
{
return rescale_count ;
}

#ifdef GEN_HUGE_FUNCS
HugeInline void
Context::ConvertToOrdinaryArray( void )
{
#ifndef NO_FAST_STATS
    int 	size = GetMaxIndex() ;
    int 	i ;

#ifdef STATS
Model2::active_model->ToFastConversions++ ;
#endif
if( ! IsLeaf() ){
	Links	    *new_links = Model2::NewLinks( CharsCount, &links ) ;

    memset( new_links, 0, sizeof( Links ) * CharsCount ) ;
    for( i = 0 ; i <= size ; i++ )
	new_links[ (int)stats[ i ].symbol ].next = links[ i ].next ;

    Model2::DeleteLinks( links ) ;
    links = new_links ;
    }

    {
	FStats	   *new_stats = Model2::NewFastStats( &fstats ) ;

    memset( new_stats, 0, sizeof( FStats ) * CharsCount ) ;
    for( i = 0 ; i <= size ; i++ )
	new_stats[ (int)stats[ i ].symbol ].count = stats[ i ].count ;

    Model2::DeleteStats( stats ) ;
    fstats = new_stats ;
    }

max_index = CharsCount - 1 ;
flags	 |= ContextFast ;
#else
cerr << "Context::ConvertToOrdinaryArray() called\n" ;
exit( -1 ) ;
#endif
}

HugeInline void
Context::ConvertToLooseArray( void )
{
#ifndef NO_FAST_STATS
static
    ScoreBoard	     used ;	// To avoid undue initialization...
    int 	     i ;

used.Clear() ;	// It is _much_ faster then constructor call...
#ifdef STATS
Model2::active_model->ToSlowConversions++ ;
#endif
if( IsLeaf() ){
    for( i = 0 ; i < CharsCount ; i++ )
	if( fstats[ i ].count > 0 )
	    used.Set( CharType( i ) ) ;
    }
else {
    for( i = 0 ; i < CharsCount ; i++ )
	if( fstats[ i ].count > 0 || links[ i ].next != NULL )
	    used.Set( CharType( i ) ) ;
    }

if( ! IsLeaf() ){
	Links	   *new_links = Model2::NewLinks( used.CheckCount(), &links ) ;
	Links	   *lt	      = new_links ;

    for( i = 0 ; i < CharsCount ; i++ )
	if( used.Checked( CharType( i ) ) )
	    *lt++ = links[ i ] ;

    Model2::DeleteLinks( links ) ;
    links = new_links ;
    }

    {
	Stats	   *new_stats = Model2::NewStats( used.CheckCount(), &stats ) ;
	Stats	   *st	      = new_stats ;

    for( i = 0 ; i < CharsCount ; i++ )
	if( used.Checked( CharType( i ) ) ){
	    st->symbol = CharType( i ) ;
	    st->count  = fstats[ i ].count ;
	    st++ ;
	    }

    Model2::DeleteFastStats( fstats ) ;
    stats = new_stats ;
    }

max_index = used.CheckCount() - 1 ;
flags	 &= ~ContextFast ;
#else
cerr << "Context::ConvertToLooseArray() called\n" ;
exit( -1 ) ;
#endif
}
#endif

inline void
Context::IncrementUseCount( void )
{
#ifdef DEBUG
CheckSanity( "IncrementUseCount" ) ;
#endif
use_count++ ;
}

inline unsigned long
Context::GetUseCount( void )
{
#ifdef DEBUG
CheckSanity( "GetUseCount" ) ;
#endif
return use_count ;
}

inline void
Context::ScaleUseCount( void )
{
#ifdef DEBUG
CheckSanity( "ScaleUseCount" ) ;
#endif
use_count /= ModelWeightsReduce ;
}

inline Context *
Context::GetNext( int index ) const
{
#ifdef	DEBUG
CheckSanity( "GetNext" ) ;
if( IsLeaf() ){
    cerr << "Context::GetNext sanity check: called on Leaf node\n" ;
    exit( -1 ) ;
    }
#endif
return index < 0 ? NULL : links[ index ].next ;
}

inline Context *
Context::GetNext( CharType c ) const
{
return GetNext( CharToIndex( c ) ) ;
}

inline void
Context::SetNext( int index, Context *next )
{
#ifdef	DEBUG
CheckSanity( "SetNext" ) ;
if( IsLeaf() ){
    cerr << "Context::SetNext sanity check: called on Leaf node\n" ;
    exit( -1 ) ;
    }
if( index < 0 || index > GetMaxIndex() ){
    cerr << "Context::SetNext sanity check: link outside of range\n" ;
    exit( -1 ) ;
    }
#endif
links[ index ].next = next ;
}

inline void
Context::SetNext( CharType c, Context *next )
{
SetNext( MakeIndex( c ), next ) ;
}

inline CountType
Context::GetCount( int index ) const
{
#ifdef DEBUG
CheckSanity( "GetCount" ) ;
#endif
return index < 0 ? CountType( 0 ) : FastStats() ? fstats[ index ].count : stats[ index ].count ;
}

inline CountType
Context::GetCount( CharType c ) const
{
return GetCount( CharToIndex( c ) ) ;
}

inline void
Context::SetCount( int index, CountType count )
{
#ifdef DEBUG
CheckSanity( "SetCount" ) ;
if( index < 0 || index > GetMaxIndex() ){
    cerr << "Context::SetCount sanity check: index outside of bounds\n" ;
    exit( -1 ) ;
    }
#endif
FastStats() ? fstats[ index ].count : stats[ index ].count = count ;
}

inline void
Context::IncrementCharCount( CharType c )
{
    int 	index	  = MakeIndex( c ) ;
    CountType	new_count = GetCount( index ) ;

#ifdef CODE_BURST
if( IsBurstable() ){
#if 0
    //
    // Post-burst - abandon it: though better in a worst-case, it gives
    // next to nothing on the normal text...
    //
    if( burst_char == c )
	new_count += CodeBurstValue-1 ;
    burst_char = NoBurstChar ;
    if( new_count == 0 )
	burst_char = c ;
#else
    //
    // Pre-burst - nice thing it is ;)
    //
    if( burst_char != NoBurstChar ){
	if( CharType( burst_char ) != c ){
	    //
	    //	Burst was a bad idea, undo it...
	    //
		int	    burst_index = CharToIndex( burst_char ) ;

	    if( burst_index >= 0 ){
		CountType   burst_count = GetCount( burst_index ) ;

		if( burst_count >= CodeBurstValue )
		    SetCount( burst_index, burst_count - (CodeBurstValue-1) ) ;
		}
	    }
	else ; // Burst was a good idea, approve it...
	burst_char = NoBurstChar ;
	}
    if( new_count == 0 ){
	//
	//  We see this symbol for the first time, burst it...
	//
	new_count  += (CodeBurstValue-1) ;
	burst_char  = c ;
	}
#endif
    }
#endif
new_count++ ;
SetCount( index, new_count ) ;
if( new_count >= CountTypeMax ){
    #ifdef STATS
    Model2::active_model->CharCountOverflows++ ;
    #endif
    RescaleCharCounts() ;
    }
}

inline Boolean
Context::IsLeaf( void ) const
{
#ifdef DEBUG
CheckSanity( "IsLeaf" ) ;
#endif
return flags & ContextLeaf ? True : False ;
}

inline Boolean
Context::IsBurstable( void ) const
{
return flags & ContextBurstable ? True : False ;
}

inline void
Context::SetBurstableFlag( void )
{
flags |= ContextBurstable ;
}

inline int
Context::GetMaxIndex( void ) const
{
#ifdef DEBUG
CheckSanity( "GetMaxIndex" ) ;
#endif
return max_index ;
}

inline CharType
Context::IndexToChar( int index ) const
{
#ifdef DEBUG
CheckSanity( "IndexToChar" ) ;
if( index < 0 || index > GetMaxIndex() ){
    cerr << "Context::IndexToChar sanity check: index outside of bounds\n" ;
    exit( -1 ) ;
    }
#endif
return FastStats() ? CharType( index ) : stats[ index ].symbol ;
}

inline int
Context::CharToIndex( CharType key ) const
{
    int 	low, high, middle ;
    CharType	val ;

#ifdef DEBUG
CheckSanity( "CharToIndex" ) ;
#endif
//
//  If character table is not in loose-array format, just return char itself...
//
#ifndef NO_FAST_STATS
if( FastStats() )
    return int( key ) ;
#endif
//
//  We assume what stats[] is sorted by symbol value,
//  so that we can use binary search algorithm.
//
if( max_index < 0 ) return -1 ;
for( low = 0, high = max_index ; high - low > 1 ; ){
    middle = ( low + high ) / 2 ;
    if( ( val = stats[ middle ].symbol ) == key )
	return middle ;
    if( val < key )
	 low  = middle ;
    else high = middle ;
    }
if( stats[ low	].symbol == key ) return low ;
if( stats[ high ].symbol == key ) return high ;
return -1 ;
}

inline int
Context::MakeIndex( CharType key )
{
    int     insertion_point = -1 ;

#ifdef DEBUG
CheckSanity( "MakeIndex" ) ;
#endif
#ifndef NO_FAST_STATS
if( FastStats() )
    return int( key ) ;
#endif
//
//  We assume what stats[] is sorted by symbol value,
//  so that we can use binary search algorithm.
//
do {
	int	    low, high, middle ;
	CharType    val ;

    if( max_index < 0 ) break ;
    for( low = 0, high = max_index ; high - low > 1 ; ){
	middle = ( low + high ) / 2 ;
	if( ( val = stats[ middle ].symbol ) == key )
	    return middle ;
	if( val < key )
	     low  = middle ;
	else high = middle ;
	}
    if( stats[ low  ].symbol == key ) return low ;
    if( stats[ high ].symbol == key ) return high ;
    insertion_point = high ;
    if( stats[ insertion_point ].symbol > key ){
	insertion_point-- ;
	if( insertion_point >= low )
	    if( stats[ insertion_point ].symbol > key )
		insertion_point-- ;
	}
    } while( 0 ) ;

return CreateIndex( key, insertion_point ) ;
}

#ifdef GEN_HUGE_FUNCS
HugeInline int
Context::CreateIndex( CharType key, int insertion_point )
{
#ifdef DEBUG
if( FastStats() ){
    cerr << "Context::CreateIndex(): method not defined for fast stats\n" ;
    exit( -1 ) ;
    }
#endif
#ifndef NO_FAST_STATS
if( ! GoodForLoose( GetMaxIndex() + 1 ) ){
    //
    //	If we have enoght characters in this context, let's convert it
    //	from loose array into ordinary array.
    //
    ConvertToOrdinaryArray() ;
    return int( key ) ;
    }
#endif

    int 	i ;
    int 	old_max_index	= max_index++ ;

    {
	Stats	*new_stats = Model2::NewStats( GetMaxIndex() + 1, &stats ) ;
	Stats	*sf = stats + old_max_index ;
	Stats	*st = new_stats + GetMaxIndex() ;

    for( i = old_max_index ; i > insertion_point ; i-- )
	*st-- = *sf-- ;
    st->symbol = key ;
    st->count  = 0 ;
    st-- ;
    if( sf != st )
	for( ; i >= 0 ; i-- )
	    *st-- = *sf-- ;
    Model2::DeleteStats( stats ) ;
    stats = new_stats ;
    }

if( ! IsLeaf() ){
	Links	*new_links = Model2::NewLinks( GetMaxIndex() + 1, &links ) ;
	Links	*lf = links + old_max_index ;
	Links	*lt = new_links + GetMaxIndex() ;

    for( i = old_max_index ; i > insertion_point ; i-- )
	*lt-- = *lf-- ;
    lt--->next = NULL ;
    if( lf != lt )
	for( ; i >= 0 ; i-- )
	    *lt-- = *lf-- ;
    Model2::DeleteLinks( links ) ;
    links = new_links ;
    }

return insertion_point + 1 ;
}

HugeInline void
Context::RescaleCharCounts( void )
{
    int 	i, size ;
    CountType	new_count ;
    int 	zero_counts	 = 0 ;
    Boolean	was_in_fast_form = FastStats() ;

#ifdef STATS
Model2::active_model->Rescalings++ ;
#endif
#ifdef DEBUG
CheckSanity( "RescaleCharCounts" ) ;
#endif
if( rescale_count < CHAR_MAX )
    rescale_count ++ ;
for( i = 0, size = GetMaxIndex() ; i <= size ; i++ ){
    if( ( new_count = GetCount( i ) / 2 ) == 0 )
	zero_counts++ ;
    SetCount( i, new_count ) ;
    }
if( IsLeaf() && zero_counts > 0 ){
    //
    //	For the leaf contexts, we can reduce amount of occupied memory
    //	by freeing empty counts
    //
	int	    new_max_index = GetMaxIndex() - zero_counts ;

    if( new_max_index >= 0 ){
	if( was_in_fast_form )
	    if( GoodForLoose( new_max_index ) )
		ConvertToLooseArray() ;
	    else ; // Not a loose array - can't compact it.
	else {

		Stats	   *new_stats = Model2::NewStats( new_max_index + 1, &stats ) ;
		Stats	   *f = stats, *t = new_stats ;
		int	    i = GetMaxIndex() ;

	    while( i-- >= 0 )
		if( f->count > 0 )
		      *t++ = *f++ ;
		else  f++ ;

	    Model2::DeleteStats( stats ) ;
	    stats     = new_stats ;
	    max_index = new_max_index ;
	    }
	}
    else {
	Model2::DeleteStats( stats ) ;
	stats	   = NULL ;
	flags	  &= ~ContextFast ;
	max_index  = new_max_index ;
	}
    }
}

HugeInline void
Context::KillEmptyLinks( void )
{
//
//  Compact this node links, if possible
//
static
    ScoreBoard	    empties ;
    int 	    i, nodes ;
    int 	    size = GetMaxIndex() ;
    Boolean	    was_in_fast_form = FastStats() ;

empties.Clear() ;	// It is _much_ faster then constructor call...

for( i = 0, nodes = 0 ; i <= size ; i++ )
    if( GetNext( i ) == NULL && GetCount( i ) == 0 ){
	nodes++ ;
	empties.Set( CharType( i ) ) ;
	}

if( nodes == 0 )
    return ;

    int 	    new_size = size - nodes ;

if( was_in_fast_form )
    if( GoodForLoose( new_size ) )
	ConvertToLooseArray() ;
    else ; // Can't compact ordinary array
else {
    //
    //	I can't merge two following cycles, because second
    //	call to the memory allocator could cause garbage collection,
    //	which will make pointers to the first block invalid.
    //
	{
	    Stats      *new_stats = Model2::NewStats( new_size + 1, &stats ) ;
	    Stats      *sf = stats, *st = new_stats ;

	for( i = 0 ; i <= size ; i++, sf++ )
	    if( ! empties.Checked( CharType( i ) ) )
		*st++ = *sf ;
	Model2::DeleteStats( stats ) ;
	stats = new_stats ;
	}

	{
	    Links      *new_links = Model2::NewLinks( new_size + 1, &links ) ;
	    Links      *lf = links, *lt = new_links ;

	for( i = 0 ; i <= size ; i++, lf++ )
	    if( ! empties.Checked( CharType( i ) ) )
		*lt++ = *lf ;
	Model2::DeleteLinks( links ) ;
	links = new_links ;
	}
    max_index = new_size ;
    }
}
#endif

inline void
Context::DeleteLeastUsed( void )
{
    int 	    i, deleted ;
    int 	    size = GetMaxIndex() ;

#ifdef DEBUG
CheckSanity( "DeleteLeastUsed" ) ;
#endif

RescaleCharCounts() ;	// ??? May be commening this out could increase compression ratio?
			// Yes, it does - for the text files. But it decreases ratio for
			// binary ones...
for( i = 0, deleted = 0 ; i <= size ; i++ ){
	Context     *c = GetNext( i ) ;

    if( c != NULL && c->GetUseCount() < ModelWeightsReduce ){
	SetNext( i, NULL ) ;
	delete c ;
	deleted++ ;
	}
    }

if( deleted )
    KillEmptyLinks() ;
}

inline void *
Context::operator new( size_t size )
{
#ifdef DEBUG
if( size != sizeof( Context ) ){
    cerr << "Context::operator new called with unexpected size " << size << "\n" ;
    exit( -1 ) ;
    }
#endif
return Model2::active_model->memory.AllocateFixed() ;
}

inline void
Context::operator delete( void *p )
{
Model2::active_model->memory.FreeFixed( p ) ;
}

