//
//  It is file 'machine.cc'
//
//  Copyright (C) 1992, 1993 Serge Pachkovsky, ps@che.nsk.su
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 1, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
#include "stdinc.hpp"
#include "timestamp.hpp"
#include "machine.hpp"
#include "crc32.hpp"
#include "codes.hpp"
#include "xio.hpp"
#include "buf_file.hpp"
#include "crcstream.hpp"
#include "byteorder.hpp"
#include "archdr.hpp"

static void
ReportProgress( unsigned long curr, unsigned long max )
{
cerr << setw( 10 ) << curr
     << " (" << setw( 3 ) << ( 100 * curr ) / max << "%)"
     << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b" << flush ;
#ifdef __GNUC__
//
//  iostream library supplied with G++ 2.2.2 will broke down if too much
//  characters without newline character was output to ostream. So, we'll
//  hack our way around this bug...
//
cerr.seekp( cerr.tellp() ) ;
#endif
}

static void
SayFileName( const char *name )
{
	int	width = ( ( strlen( name ) + 13 ) / 14 ) * 14 ;

cerr.setf( ios::left, ios::adjustfield ) ;
cerr << setw( width ) << name << " " << ends ;
cerr.setf( ios::internal, ios::adjustfield ) ;
}

static int
CountChars( BufferedFile & file )
{
	ScoreBoard	board ;
	int		c ;

while( ( c = file.GetByte() ) >= 0 )
    board.Set( CharType( c ) ) ;
file.Rewind() ;
#ifdef DEBUG
cerr << "CountChars: File have " << board.CheckCount() << "different chars\n" ;
#endif
return board.CheckCount() ;
}

CompletionCode
CompressFile( int out, char * const name, int order, unsigned long memory )
{
    BufferedFile    in( name, ModeBinRead ) ;

if( in.code != Completed )
    return in.code ;

memory = ( memory + 1023L ) & ~1023L ;

    long	    orig_position = tell( out ) ;
    UncomprHdr	    hdr1 ;

memset( &hdr1, 0, SizeOfUncomprHdr ) ;
hdr1.MemoryUsed = (unsigned short)( memory / 1024 ) ;
hdr1.ComprOrder = order ;
ToIntel( &hdr1, hdr1.Layout ) ;
if( write( out, &hdr1, SizeOfUncomprHdr ) != SizeOfUncomprHdr )
    return OutOfDisk ;
ToNative( &hdr1, hdr1.Layout ) ;

SayFileName( name ) ;
cerr << "Scanning..." ;
hdr1.CharsCount = CountChars( in ) ;
cerr << "\b\b\b\b\b\b\b\b\b\b\b" "           " "\b\b\b\b\b\b\b\b\b\b\b" ;

    ComprHdr		hdr2 ;

memset( &hdr2, 0, SizeOfComprHdr ) ;
hdr2.FileLength  = filelength( fileno( in.f ) ) ;
hdr2.NameLength  = strlen( name ) ;
if( GetTimeStamp( name, hdr2.Modified ) < 0 )
    return CantGetTimeStamps ;

    const unsigned long     store_trigger_level = hdr2.FileLength + hdr2.NameLength +
						  SizeOfComprHdr  - TriggerStoringThreshold ;
    Boolean		    incomplete ;

do {
    incomplete = False ;
	{
	    CompStream	    stream( out, order, memory, hdr1.CharsCount ) ;
	    int 	    c ;

//	cerr << "Length before ToIntel: " << hdr2.FileLength ;
	ToIntel( &hdr2, hdr2.Layout ) ;
	stream.WriteBuf( SizeOfComprHdr,  (BYTE *)&hdr2 ) ;
	ToNative( &hdr2, hdr2.Layout ) ;
//	cerr << "Length after ToNative: " << hdr2.FileLength ;
	stream.WriteBuf( hdr2.NameLength, (BYTE *)name ) ;

	for( unsigned long i = 0 ; i < hdr2.FileLength ; i++ ){
	    if( ( c = in.GetByte() ) < 0 ){
		perror( name ) ;
		return ReadError ;
		}
	    stream.WriteByte( c ) ;
	    if( stream.GetCompressedSize() >= store_trigger_level ){
		//
		//  Ok, this file could not be compressed and we'll have to store it.
		//  So, there is no sense in fiddling aroung with the rest of file.
		//
		incomplete	= True ;
		break ;
		}
	    if( ( i & (ProgressReportStep-1) ) == 0 )
		ReportProgress( i, hdr2.FileLength ) ;
	    }
	hdr1.CRC32	= stream.crc ;
	hdr1.MemoryUsed = ( stream.str->GetMaxMemoryUsed() + 1023 ) / 1024 ;
	}
    //
    //	At this point, compression stream is closed, and it is safe to fool around
    //	with I/O offset...
    //
    if( incomplete ){
	cerr << "Storing ... " ;
	lseek( out, orig_position + SizeOfUncomprHdr, SEEK_SET ) ;
	hdr1.ComprOrder =
	order		= FileIsStored ;
	in.Rewind() ;
	}
    } while( incomplete ) ;

    long	    end_position = tell( out ) ;

lseek( out, orig_position, SEEK_SET ) ;
hdr1.BlockSize = end_position - orig_position ;
ToIntel( &hdr1, hdr1.Layout ) ;
if( write( out, &hdr1, SizeOfUncomprHdr ) != SizeOfUncomprHdr )
    return OutOfDisk ;
ToNative( &hdr1, hdr1.Layout ) ;

    unsigned	    percent	 = unsigned( ( hdr1.BlockSize * 1000L ) / hdr2.FileLength ) ;

lseek( out, orig_position, SEEK_SET ) ;
cerr << "Compressed to " ;
cerr << setw( 8 ) << hdr1.BlockSize << " bytes "
     << "(" << setw( 3 ) << percent / 10 << "." << percent % 10 << "%)" << endl ;

return Completed ;
}

CompletionCode
DecompressFile( int in, FileInfo *info, DecompressActions action )
{
    long	    orig_position = tell( in ) ;
    UncomprHdr	    hdr1 ;

if( read( in, &hdr1, SizeOfUncomprHdr ) != SizeOfUncomprHdr ){
    perror( "Can't read compressed file header" ) ;
    return ReadError ;
    }
ToNative( &hdr1, hdr1.Layout ) ;
info->CompSize	 = hdr1.BlockSize ;
info->CRC32	 = hdr1.CRC32 ;
info->MemoryUsed = hdr1.MemoryUsed * 1024L ;
info->ModelOrder = hdr1.ComprOrder ;
info->CharsCount = hdr1.CharsCount ;
info->name	 = new char[ MaxNameLength + 1 ] ;

    unsigned long   memory_size = DefaultMemorySize ;

if( action != GetInfo && info->MemoryUsed > DefaultMemorySize / 2 ){
    const unsigned long overhead   = 2 * (DefaultBitStreamBuffer + IOBufSize) +
				     MemoryPool::AllocatorOverhead( 3 * info->MemoryUsed / 2 ) ;
    const unsigned long needed_mem = info->MemoryUsed + overhead ;
//  const unsigned long phys_mem   = AvailablePhysicalMemory( needed_mem ) ;
    const unsigned long virt_mem   = AvailableVirtualMemory ( needed_mem ) ;

    //
    //	First, let's forget about this cruel world: allocated all we could
    //	need from virtual memory. If were is not enougth physical memory,
    //	we'll complaint about it later.
    //
    if( virt_mem >= needed_mem ){
	memory_size = 2 * (virt_mem - needed_mem) / 3 ;
	if( memory_size > info->MemoryUsed / 2 )
	    memory_size = info->MemoryUsed / 2 ;

	memory_size += info->MemoryUsed ;
	}
    else
	memory_size = virt_mem - overhead ;
    //
    //	Somehow, I don't think we should care about fitting into real memory
    //	if it is big enougth to accomodate our bare necessities but far less
    //	then we could use in virtual space. It is a tradeoff between causing
    //	GC to run more often and experiencing a slight swapping...
    //
    if( memory_size < info->MemoryUsed )
	cerr << "(Memory low) " << flush ;
    }

    DecompStream    stream( in, info->ModelOrder, memory_size, info->CharsCount ) ;
    ComprHdr	    hdr2 ;

stream.ReadBuf( SizeOfComprHdr, (BYTE *)&hdr2 ) ;
ToNative( &hdr2, hdr2.Layout ) ;

info->UncompSize = hdr2.FileLength ;
info->Modified	 = hdr2.Modified ;
if( hdr2.NameLength > MaxNameLength ){
    //
    //	Memory fragmentation kludge: allocation of filename buffer at this
    //	time will split memory into two fragments, thus causing problems
    //	for DOS version with limited memory available.
    //
    delete info->name ;
    info->name	 = new char[ hdr2.NameLength + 1 ] ;
    }
stream.ReadBuf( hdr2.NameLength, (BYTE *)info->name ) ;
info->name[ hdr2.NameLength ] = 0 ;
CanonicalizeName( info->name ) ;
if( action == GetInfo ){
    lseek( in, orig_position, SEEK_SET ) ;
    return Completed ;
    }
//
//  And now, decompress file!
//
SayFileName( info->name ) ;

    const char *    true_name = ( action == TestFragment ? NullDevice : info->name ) ;
    BufferedFile    out( true_name, ModeBinWrite ) ;

if( out.code != Completed )
    return out.code ;

    unsigned long   i ;
    int 	    c ;

    for( i = 0 ; i < hdr2.FileLength ; i++ ){
	if( ( c = stream.ReadByte() ) == EOF ){
	    perror( info->name ) ;
	    return ReadError ;
	    }
	#ifdef TRACE_CODES
	cout << ":: " << i << " = " << c << "\n" ;
	#endif
	if( out.PutByte( c ) == EOF ){
	    perror( info->name ) ;
	    return WriteError ;
	    }
	if( ( i & (ProgressReportStep-1) ) == 0 )
	    ReportProgress( i, hdr2.FileLength ) ;
	}

lseek( in, orig_position, SEEK_SET ) ;

out.Close() ;
SetTimeStamp( true_name, hdr2.Modified ) ;

if( stream.crc == hdr1.CRC32 ){
    cerr << "Decompressed        " << endl ;
    return Completed ;
    }
else {
    cerr << "File CRC do not match!" << endl ;
    return BadCRC ;
    }
}

void
FileInfo::Title( ostream &str )
{
str << "   File name       Used      Size    Last Modified   Order  Memory   CharSet" << endl ;
}

ostream &
operator<<( ostream &str, FileInfo &info )
{
str.setf( ios::left, ios::adjustfield ) ;
if( strlen( info.name ) > 14 )
    str << info.name << "\n               " ;
else
    str << setw( 14 ) << info.name << " " ;
str.setf( ios::internal, ios::adjustfield ) ;
str << setw( 8 ) << info.CompSize   << " ("
    << setw( 8 ) << info.UncompSize << ") "
    << info.Modified		    << "   " ;
if( info.ModelOrder != FileIsStored ){
    str << setw( 2 ) << info.ModelOrder << "    "
	<< setw( 4 ) ;
    if( info.MemoryUsed >= 9999L * 1024 )
	 str << info.MemoryUsed / (1024L*1024) << "M" ;
    else str << info.MemoryUsed / (1024L)      << "K" ;
    str << "    " ;
    str << setw( 3 ) ;
    str << info.CharsCount ;
    }
else
    str << "(stored)" ;
return str ;
}

const char *
MachineStrerror( CompletionCode errno )
{
switch( errno ){
    case Completed:	    return "No errors during decompression" ;
    case GenericError:	    return "Undefined engine error" ;
    case OutOfMemory:	    return "Out of memory in engine" ;
    case OutOfDisk:	    return "Out of disk space in engine" ;
    case Break: 	    return "Break" ;
    case FileNotFound:	    return "File not found" ;
    case ReadError:	    return "File read error" ;
    case WriteError:	    return "File write error" ;
    case BadCRC:	    return "File CRC do not match" ;
    case CantGetTimeStamps: return "Can't get time stamps" ;
    case CantSetTimeStamps: return "Can't set time stamps" ;
    default:		    return "Invalid engine error code" ;
    }
}
